#include "manager.h"
#include <stdio.h>

haManager* haManager::chain;

const int haManager::blockChain[32] = { 
     0,  0,  1,  2,  3,  4,  5,  6,  7,  8, 
     9,  9, 10, 10, 10, 10, 11, 11, 11, 11, 
    11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 
    12, 12
};

const int haManager::blockSize[32] = { 
     16,  16,  24,  32,  40,  48,  56,  64,  72,  80,  
     96,  96, 128, 128, 128, 128, 168, 168, 168, 168, 
    168, 256, 256, 256, 256, 256, 256, 256, 256, 256, 
    256, 256
};

#define ALIGN(x, p) (((x) + (p) - 1) & ~((p)-1))

bool haManager::open(char const* name, char const* host, int localPort, int remotePort, 
		     void* baseAddr)
{
    int size;
    port = localPort;
    haSocket* s = haSocket::connect(host, remotePort);
    shmem.open(name, maxSize, baseAddr);
    initialize();
    if (s == NULL) { // master
	printf("Active node started\n");
	s = new haSocket(localPort);
	prevStandbySock = standbySock = s->accept();
	standbySock->next = NULL;
	size = 0;
	standbySock->write(&size, sizeof size);
	shmem.writeProtect();
    } else { 
	printf("Standby start sync\n");
	if (s->read(&size, sizeof size)
	    && (size == 0 || s->read(shmem.base(), size)))
	{
	    printf("Standby node ready\n");
	    int bufSize = 64*(1024+4);
	    char* buf = (char*)malloc(bufSize);
	    while (s->read(&size, sizeof size)) { 
		if (size < 0) { 
		    delete s;
		    free(buf);
		    return false;
		}		
		if (size > bufSize) { 
		    buf = (char*)realloc(buf, size); 
		    bufSize = size;
		}
		char* src = buf;
		char* end = src + size;
		if (!s->read(src, size)) { 
		    break;
		}
		while (src < end) { 
		    int offs = *(int*)src;
		    src += sizeof(int);
		    memcpy(shmem.base() + offs, src, pageSize);
		    src += pageSize;
		}
	    }
	    free(buf);
	}
	printf("Active node failure detected\n");
	s = NULL;
	currFreePage = hdr->freePageChain;
    }
    delete s;
    acceptThread.run(acceptProc, this);
    updateThread.run(updateProc, this);
    next = chain;
    chain = this;
    return true;
}


void THREAD_PROC haManager::acceptProc(void* arg)
{
    ((haManager*)arg)->recovery();
}


void THREAD_PROC haManager::updateProc(void* arg)
{
    ((haManager*)arg)->update();
}
    
void haManager::recovery()
{    
    haSocket* s = new haSocket(port);
    while (acceptConnections) { 
	haSocket* newSock = s->accept();
	printf("Standby node reconnected\n");
	newSock->next = standbySock;
	standbySock = newSock;
    }
    delete s;
}

void haManager::update()
{
    haPage* pg;
    while ((pg = queue.get()) != NULL) { 
	bool rc;
	haSocket* s = standbySock;
	if (s != NULL && prevStandbySock == s) { 
	    if (pg->size != 0) { 
		rc = s->write(&pg->size, pageSize+8);
	    } else {  
		rc = s->write(&pg->offs, pageSize+4);
	    }
	    if (!rc) { 
		prevStandbySock = NULL;
	    }
	}
	queue.free(pg);	    
    }
}


bool haManager::pageFaultHandler(char* address)
{
    haManager* mng = getManager(address);
    if (mng != NULL) {
	mng->handlePageFault(address - mng->shmem.base());
	return true;
    }
    return false;
}


haManager::haManager(size_t size)
: queue(shmem.getPageSize())
{
    maxSize = size;
    dirtyPageMap = NULL;
}

haObjectHeader* haManager::getHeader(void* obj) 
{ 
    haObjectHeader* hp;
    size_t addr = size_t(obj) - sizeof(haObjectHeader);
    size_t offs = addr - size_t(hdr);
    int pageNo = offs / allocPageSize;

    if (hdr->pageMap == NULL 
	|| !(hdr->pageMap->bits[pageNo >> 3] & (1 << (pageNo & 7))))
    { 
	hp = (haObjectHeader*)(addr & ~(allocPageSize-1));
	if (hp->size + sizeof(haObjectHeader) <= allocPageSize/2) { 
	    size_t size = 
		blockSize[((hp->size + sizeof(haObjectHeader)+7) >> 3) - 1];
	    if (size & (size-1)) { 
		offs &= allocPageSize-1; //offset within page
		hp = (haObjectHeader*)(addr - (offs % size));
	    } else { // power of 2
		hp = (haObjectHeader*)(addr & ~(size-1));
	    }
	} 
    } else { // large object occupies several pages
	do { 
	    pageNo -= 1;
	} while (hdr->pageMap->bits[pageNo >> 3] & (1 << (pageNo & 7)));
	hp = (haObjectHeader*)((char*)hdr + pageNo*allocPageSize);
    }
    assert(hp->cid != haObjectHeader::CID_FREE);
    return hp;
}

void* haManager::allocPage(int cid, size_t size)
{
    haStorageFreeBlock* bp = NULL;
    size_t alignedSize = ALIGN(size + sizeof(haObjectHeader), allocPageSize);
    if (hdr->freePageChain != NULL) { 
	haStorageFreeBlock** bpp = &currFreePage->next;
	while ((bp = *bpp) != NULL && bp->size < size) { 
	    bpp = &bp->next;
	}
	if (bp == NULL) { 
	    bpp = &hdr->freePageChain;
	    haStorageFreeBlock* saveNext = currFreePage->next;
	    currFreePage->next = NULL;
	    while ((bp = *bpp) != NULL && bp->size < size) { 
		bpp = &bp->next;
	    }
	    currFreePage->next = saveNext;
	}
	if (bp != NULL) { 
	    if (bp->size > alignedSize) { 
		currFreePage = bp;
		bp->size -= alignedSize;
		bp = (haStorageFreeBlock*)
		    ((char*)bp + bp->size + sizeof(haObjectHeader));
	    } else { 
		*bpp = bp->next;
		currFreePage = bp->next ? bp->next : hdr->freePageChain;
	    } 	    
	} 
    } 
    if (bp == NULL) { 
	bp = (haStorageFreeBlock*)((char*)hdr + hdr->usedSize);
	hdr->usedSize += alignedSize;
	if (hdr->pageMap != NULL && cid != haObjectHeader::CID_BITMAP) { 
	    size_t oldMapSize = hdr->pageMap->size;
	    if (hdr->usedSize > oldMapSize*allocPageSize*8) {
		// fileSize+mapSize+mapHeader >= pageSize*8*mapSize
		size_t newMapSize = 
		    (hdr->usedSize + sizeof(haObjectHeader) + allocPageSize*8-2)
 		                       / (allocPageSize*8 - 1);
		// reserve space for future
		newMapSize = ALIGN(newMapSize + sizeof(haObjectHeader),
				   allocPageSize)*2 - sizeof(haObjectHeader); 
		haStoragePageMap* newMap = (haStoragePageMap*)
		    ((char*)allocPage(haObjectHeader::CID_BITMAP, newMapSize) - sizeof(haObjectHeader));
		memcpy(newMap->bits, hdr->pageMap->bits, oldMapSize);
		memset(newMap->bits + oldMapSize, 0, 
		       newMapSize - oldMapSize);
		freePage(hdr->pageMap);
		hdr->pageMap = newMap;
	    }
	} 
    }
    bp->cid = cid;
    bp->size = size;
    if (size + sizeof(haObjectHeader) <= allocPageSize/2) { 
        int n = ((size + sizeof(haObjectHeader) + 7) >> 3) - 1;
	size = blockSize[n];
	haStorageFreeBlock* endOfPage = 
	    (haStorageFreeBlock*)((char*)bp + allocPageSize - size + 1);
	haStorageFreeBlock* cp = (haStorageFreeBlock*)((char*)bp + size);
	do { 
	    cp = cp->next = (haStorageFreeBlock*)((char*)cp + size);
	} while (cp < endOfPage);
	cp = (haStorageFreeBlock*)((char*)cp - size);
	cp->next = hdr->freeBlockChain[blockChain[n]];
	hdr->freeBlockChain[blockChain[n]] = 
	    (haStorageFreeBlock*)((char*)bp+size);
    } else if (alignedSize > allocPageSize && cid != haObjectHeader::CID_BITMAP) {
	if (hdr->pageMap == NULL) { 
	    // fileSize+mapSize+mapHeader >= pageSize*8*mapSize
	    size_t mapSize = (shmem.size() + sizeof(haObjectHeader)+allocPageSize*8-2)
		/ (allocPageSize*8 - 1);
	    mapSize = ALIGN(mapSize + sizeof(haObjectHeader), allocPageSize)
		* initPageMapSize - sizeof(haObjectHeader); 
	    hdr->pageMap = (haStoragePageMap*)((char*)allocPage(mapSize, true) - sizeof(haObjectHeader));
	    memset(hdr->pageMap->bits, 0, mapSize);
	}
	int pageNo = size_t((char*)bp - (char*)hdr) / allocPageSize; 
	int nPages = alignedSize / allocPageSize - 1;
	do { 
	    pageNo += 1;
	    hdr->pageMap->bits[pageNo >> 3] |= 1 << (pageNo & 7);
	} while (--nPages != 0);
    } 
    return (haObjectHeader*)bp+1;
}



void haManager::freePage(haObjectHeader* hp)
{
    haStorageFreeBlock *bp, *prev;
    haStorageFreeBlock *fp = (haStorageFreeBlock*)hp;

    if (currFreePage != NULL && currFreePage < fp) { 
	prev = currFreePage;
	bp = prev->next;
    } else { 
	prev = NULL;
	bp = hdr->freePageChain; 
    }

    while (bp != NULL && bp < fp) { 
	 prev = bp;
	 bp = bp->next;
    }
    assert(bp != fp/*object can't be deallocated twice*/);

    size_t alignedSize = ALIGN(fp->size + sizeof(haObjectHeader), allocPageSize);
    fp->size = alignedSize - sizeof(haObjectHeader);
    if (alignedSize > allocPageSize
	&& fp->cid != haObjectHeader::CID_FREE // reconstruction of free list
	&& fp->cid != haObjectHeader::CID_BITMAP) // allocation of page map
    {
	assert(hdr->pageMap != NULL);
	int pageNo = size_t((char*)fp - (char*)hdr) / allocPageSize; 
	int nPages = alignedSize / allocPageSize - 1;
	do { 
	    pageNo += 1;
	    hdr->pageMap->bits[pageNo >> 3] &= ~(1 << (pageNo & 7));
	} while (--nPages != 0);
    }
    fp->cid = haObjectHeader::CID_FREE;
    fp->next = bp;

    if (prev == NULL) { 
	hdr->freePageChain = fp;
    } else { 
	if ((char*)prev + prev->size + sizeof(haObjectHeader) == (char*)fp) {
	    prev->size += fp->size + sizeof(haObjectHeader);
	    fp = prev;
	} else { 
	    prev->next = fp;
	}
    }
    if ((char*)fp + fp->size + sizeof(haObjectHeader) == (char*)bp) { 
	fp->size += bp->size + sizeof(haObjectHeader);
	fp->next = bp->next;
    }
    currFreePage = fp;
}

void haManager::initialize()
{
    pageSize = shmem.getPageSize();
    int mapSize = maxSize/pageSize/32;
    dirtyPageMap = (int*)malloc(mapSize*sizeof(int));
    memset(dirtyPageMap, 0, mapSize*sizeof(int));
    prevStandbySock = standbySock = NULL;
    hdr = (haStorageHeader*)shmem.base();
    if (hdr->baseAddr != (char*)hdr) { 
	memset(hdr, 0, allocPageSize);
	hdr->baseAddr = (char*)hdr;
	hdr->usedSize = allocPageSize;
    }
    currFreePage = hdr->freePageChain;
    acceptConnections = true;
}


void haManager::handlePageFault(size_t offs)
{
    int i = offs/pageSize;
    assert(offs < maxSize);
    assert((dirtyPageMap[i / 32] & (1 << (i & 31))) == 0);
    dirtyPageMap[i / 32] |= 1 << (i & 31);
    nDirtyPages += 1;
    shmem.writeUnprotect(offs & ~(pageSize-1));
}

void haManager::commit() 
{
    int n = shmem.size() / pageSize / 32;
    int* map = dirtyPageMap;
    haSocket* s = standbySock;
    if (s == prevStandbySock && s != NULL) {
	if (nDirtyPages > 0) { 
	    int i = 0;
	    for (int j = 0; j < n; j++) { 
		int mask = map[j];
		if (mask != 0) { 
		    map[j] = 0;
		    for (int k = 32; mask != 0 && --k >= 0;) { 
			if (mask < 0) { 
			    size_t offs = (j*32+k)*pageSize;
			    queue.put(i++, nDirtyPages, offs, shmem.base() + offs);
			    shmem.writeProtect(offs);
			}
			mask <<= 1;
		    }
		}
	    }
	    assert(i == nDirtyPages);
	    nDirtyPages = 0;
	}
    } else if (s != NULL) { 
	queue.clear();
	if (s->write(&hdr->usedSize, sizeof hdr->usedSize)
	    && s->write(shmem.base(), hdr->usedSize))
	{
	    prevStandbySock = s;
	} 	    
	haSocket* next = s->next;
	s->next = NULL;
	while (next != NULL) { 
	    s = next;
	    next = s->next;
	    delete s;
	}
	shmem.writeProtect();
	memset(dirtyPageMap, 0, maxSize/pageSize/8);	
	nDirtyPages = 0;
    }
}

void haManager::close()
{
    haManager** mpp;
    for (mpp = &chain; *mpp != this; mpp = &(*mpp)->next);
    *mpp = next;

    commit();
    shmem.close();
    queue.stop();
    updateThread.join();

    // notify standby node that we are closing
    haSocket* s = standbySock;
    if (s != NULL) { 
	int size = -1;
	s->write(&size, sizeof size);
    }

    // wakeup blocked accept thread
    acceptConnections = false;
    s = haSocket::connect("localhost", port);
    delete s;
    acceptThread.join();

    // Cleanup sockets
    s = standbySock;
    while (s != NULL) { 
	haSocket* next = s->next;
	delete s;
	s = next;
    }
	
    free(dirtyPageMap);
}








