#include "manager.h"

#ifdef _WIN32
static LONG WINAPI AccessViolationHandler(LPEXCEPTION_POINTERS ep)
{
    return ep->ExceptionRecord->ExceptionCode == EXCEPTION_ACCESS_VIOLATION
	&& ep->ExceptionRecord->ExceptionInformation[0]
	&& haManager::pageFaultHandler((char*)ep->ExceptionRecord->ExceptionInformation[1])
	   ? EXCEPTION_CONTINUE_EXECUTION
	   : EXCEPTION_CONTINUE_SEARCH;
}


haSharedMemory::haSharedMemory()
{
    SYSTEM_INFO sysinfo;
    GetSystemInfo(&sysinfo);
    pageSize = sysinfo.dwPageSize;
}


void haSharedMemory::open(char const* fileName, size_t size, void* addr)
{
    fd = INVALID_HANDLE_VALUE;
    baseAddr = (char*)addr;
    if (fileName != NULL) {
	fd = CreateFile(fileName, GENERIC_READ|GENERIC_WRITE, 0, 
			NULL, OPEN_ALWAYS, FILE_FLAG_RANDOM_ACCESS, NULL);
	if (fd == INVALID_HANDLE_VALUE) { 
	    throw haException("Failed to create file", GetLastError());
	}
	DWORD readBytes;
	haFileHeader hdr;
	hdr.usedSize = 0;
	hdr.baseAddr = (char*)addr;
	if (!ReadFile(fd, &hdr, sizeof hdr, &readBytes, NULL)
	    || (readBytes != 0 && readBytes != sizeof hdr))
	{ 
	    throw haException("Failed to read file header", GetLastError());
	}
	baseAddr = hdr.baseAddr;
    }
    size = (size + pageSize - 1) & ~(pageSize-1);
    fileSize = size;
    md = CreateFileMapping(fd, NULL, PAGE_READWRITE, 0, size, NULL);
    if (md == NULL) { 
	throw haException("Failed to create file mapping", GetLastError());
    }
    baseAddr = (char*)MapViewOfFileEx(md, FILE_MAP_ALL_ACCESS, 0, 0, 0, baseAddr);
    if (baseAddr == NULL) { 
	throw haException("Failed to map view of file", GetLastError());
    }
    SetUnhandledExceptionFilter(AccessViolationHandler);
}

void haSharedMemory::close()
{
    UnmapViewOfFile(baseAddr);
    CloseHandle(md);
    if (fd != INVALID_HANDLE_VALUE) { 
	CloseHandle(fd);
    }
}


void haSharedMemory::setProtection(size_t offs, size_t size, bool writeEnabled)
{
    DWORD old_prot;
    if (!VirtualProtect(baseAddr + offs, size, 
			writeEnabled ? PAGE_READWRITE : PAGE_READONLY, &old_prot)) 
    { 
	throw haException("Failed to set protection", GetLastError());
    }
}

#else // Unix

#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/resource.h>

#ifdef SA_SIGINFO
#if defined(__linux__) && !defined(__alpha__)
#undef SA_SIGINFO  // Intel Linux doesn't set sa_addr in siginfo 
#define SA_SIGINFO 0 
#else
#include <sys/siginfo.h>
#endif
#else
#if defined(__linux__) // Linux 5.0 doesn't define sigcontext 
#define sigcontext_struct sigcontext
#include <asm/sigcontext.h>
#endif
#define SA_SIGINFO 0
#endif

#ifndef SA_RESTART
#define SA_RESTART 0
#endif

#ifndef MAP_FAILED
#define MAP_FAILED (-1)
#endif

#ifndef MAP_FILE
#define MAP_FILE 0
#endif

#ifndef MAP_VARIABLE
#define MAP_VARIABLE 0
#endif

#ifndef MAP_NORESERVE
// This attrbute used in Solaris to avoid reservation space in swap file
#define MAP_NORESERVE 0
#endif

#if defined(__sun__)
extern "C" int getpagesize(void);
#endif

#if defined(__linux__) && !defined(__alpha__)
static void sigsegvHandler(int, struct sigcontext scp)
#elif defined(_AIX) || defined(__FreeBSD__)
static void sigsegvHandler(int, int, struct sigcontext *scp)
#else
static void sigsegvHandler(int, siginfo_t *info)
#endif
{
    char* faultAddr;
#if defined(__linux__) && !defined(__alpha__)
    faultAddr = (char*)scp.cr2;
#elif defined(_AIX)
    faultAddr = (char*)scp->sc_jmpbuf.jmp_context.o_vaddr;
#elif defined(__FreeBSD__)
    faultAddr = (char*)scp->sc_err;
#else
    faultAddr = (char*)info->si_addr;
#endif

    if (!haManager::pageFaultHandler(faultAddr)) {
	kill(getpid(), SIGABRT);
    }
}


haSharedMemory::haSharedMemory()
{
    pageSize = getpagesize();
}


void haSharedMemory::open(char const* fileName, size_t size, void* addr)
{
    baseAddr = (char*)addr;
    int flags = 0;
    size = (size + pageSize - 1) & ~(pageSize-1);
    fileSize = size;
    if (fileName != NULL) {
	fd = ::open(fileName, O_RDWR|O_CREAT, 0777);
	if (fd < 0) { 
	    throw haException("Failed to create file", errno);
	}
	haFileHeader hdr;
	hdr.usedSize = 0;
	hdr.baseAddr = (char*)addr;
	int readBytes = read(fd, &hdr, sizeof hdr);
	if (readBytes != 0 && readBytes != sizeof hdr) { 
	    throw haException("Failed to read file header", errno);
	}
	
        if (lseek(fd, 0, SEEK_END) < (long)size) { 
	    if (ftruncate(fd, size) != 0) { 
		throw haException("Failed to truncate file", errno);
	    }
	}
	
	baseAddr = hdr.baseAddr;
	if (baseAddr == NULL) { 
	    flags |= MAP_VARIABLE;
	}
	flags |= MAP_SHARED|MAP_FILE;
    } else { 
#ifndef MAP_ANONYMOUS 
	fd = ::open("/dev/zero", O_RDONLY, 0);
#else
	flags |= MAP_ANONYMOUS;
	fd = -1;
#endif
    }
    baseAddr = (char*)mmap(baseAddr, size, PROT_READ|PROT_WRITE, flags, fd, 0);
    if (baseAddr == (char*)MAP_FAILED) { 
	throw haException("Failed to map file", errno);
    }
    static struct sigaction sigact; 
    sigact.sa_handler = (void(*)(int))sigsegvHandler;
    sigact.sa_flags = SA_RESTART|SA_SIGINFO;
    sigaction(SIGSEGV, &sigact, NULL);
#if defined(__FreeBSD__)
    sigaction(SIGBUS, &sigact, NULL);
#endif
}

void haSharedMemory::close()
{
    munmap(baseAddr, fileSize);
    if (fd != -1) { 
	::close(fd);
    }
}


void haSharedMemory::setProtection(size_t offs, size_t size, bool writeEnabled)
{
    if (mprotect(baseAddr + offs, size, writeEnabled ? PROT_READ|PROT_WRITE : PROT_READ) != 0) 
    { 
	throw haException("Failed to set protection", errno);
    }
}

#endif














